import csv
import itertools


def _boolean(data):
    if data == "False":
        result = False
    else:
        result = True
    return result


def row_to_location(row):
    if row[4] == "0":
        sub = False
        nosub = True
    else:
        sub = True
        nosub = False
    tss = _boolean(row[6])
    term = _boolean(row[8])
    return {"have no sub-operons": nosub, "have sub-operons": sub,
            "start with tss": tss, "stop with terminator": term}


def plus_num(num_total, strain, type_):
    num_total["total"][type_] += 1
    num_total[strain][type_] += 1
    num_total["total"]["total"] += 1
    num_total[strain]["total"] += 1


def print_stat(operons, total_num, class_operon, out):
    num_features = {}
    out.write("Total number of Operons is {0}\n".format(total_num))
    out.write("The sub operon and features:\n")
    for operon in operons:
        for it in range(1, 5):
            for features in itertools.combinations(operon.keys(), it):
                check_key = 0
                for key in features:
                    if operon[key]:
                        if it == 1:
                            if key in num_features.keys():
                                num_features[key] += 1
                            else:
                                num_features[key] = 1
                        check_key += 1
                if (check_key == it) and (it != 1):
                    key = " and ".join(features)
                    if key in num_features.keys():
                        num_features[key] += 1
                    else:
                        num_features[key] = 1
    for key, value in num_features.items():
        out.write("\tthe number of operons which {0} = {1} ({2})\n".format(
                  key, value, float(value) / float(total_num)))
    out.write("mono/polycistronic:\n")
    out.write("\tno associated with genes/CDSs: {0} ({1})\n".format(
              class_operon["na"],
              float(class_operon["na"]) / float(class_operon["total"])))
    out.write("\tmonocistronic: {0} ({1})\n".format(
              class_operon["mono"],
              float(class_operon["mono"]) / float(class_operon["total"])))
    out.write("\tpolycistronic: {0} ({1})\n".format(
              class_operon["poly"],
              float(class_operon["poly"]) / float(class_operon["total"])))


def stat(input_file, out_file):
    out = open(out_file, "w")
    operons = {}
    operons_all = []
    tmp_id = ""
    f_h = open(input_file, "r")
    pre_seq_id = ""
    total_num = {}
    total_num_all = 0
    class_operon = {}
    class_operon["total"] = {"na": 0, "mono": 0, "poly": 0, "total": 0}
    for row in csv.reader(f_h, delimiter="\t"):
        if row[0] != "Operon_ID":
            if row[0] != tmp_id:
                if pre_seq_id != row[1]:
                    pre_seq_id = row[1]
                    operons[row[1]] = []
                    total_num[row[1]] = 0
                    class_operon[row[1]] = {"na": 0, "mono": 0,
                                            "poly": 0, "total": 0}
                operons[row[1]].append(row_to_location(row))
                operons_all.append(row_to_location(row))
                total_num[row[1]] += 1
                total_num_all += 1
                if row[-1] == "NA":
                    plus_num(class_operon, row[1], "na")
                elif len(row[-1].split(",")) == 1:
                    plus_num(class_operon, row[1], "mono")
                elif len(row[-1].split(",")) > 1:
                    plus_num(class_operon, row[1], "poly")
                tmp_id = row[0]
    if len(operons) > 1:
        out.write("All strains:\n")
        print_stat(operons_all, total_num_all, class_operon["total"], out)
    for strain in operons.keys():
        out.write("\n" + strain + ":\n")
        print_stat(operons[strain], total_num[strain],
                   class_operon[strain], out)
    out.close()
    f_h.close()



import os
from annogesiclib.multiparser import Multiparser
from annogesiclib.helper import Helper
from annogesiclib.detect_operon import operon
from annogesiclib.stat_operon import stat
from annogesiclib.combine_gff import combine_gff


class OperonDetection(object):
    '''detection of operon'''

    def __init__(self, args_op):
        self.multiparser = Multiparser()
        self.helper = Helper()
        self.tss_path = os.path.join(args_op.tsss, "tmp")
        self.tran_path = os.path.join(args_op.trans, "tmp")
        self.utr5_path = os.path.join(args_op.utr5s, "tmp")
        self.utr3_path = os.path.join(args_op.utr3s, "tmp")
        self.table_path = os.path.join(args_op.output_folder, "tables")
        if args_op.terms is not None:
            self._check_gff(args_op.terms, "term")
            self.term_path = os.path.join(args_op.terms, "tmp")
        else:
            self.term_path = None

    def _check_gff(self, gffs, type_):
        for gff in os.listdir(gffs):
            if gff.endswith(".gff"):
                self.helper.check_uni_attributes(os.path.join(gffs, gff))

    def _detect_operon(self, prefixs, args_op):
        for prefix in prefixs:
            out_table = os.path.join(self.table_path,
                                     "_".join([prefix, "operon.csv"]))
            print("Detecting operons of {0}".format(prefix))
            tss = self.helper.get_correct_file(
                    self.tss_path, "_TSS.gff", prefix, None, None)
            tran = self.helper.get_correct_file(
                    self.tran_path, "_transcript.gff", prefix, None, None)
            gff = self.helper.get_correct_file(
                    args_op.gffs, ".gff", prefix, None, None)
            if self.term_path is None:
                term = False
            else:
                term = self.helper.get_correct_file(
                        self.term_path, "_term.gff", prefix, None, None)
            operon(tran, tss, gff, term, args_op.tss_fuzzy,
                   args_op.term_fuzzy, args_op.length, out_table)

    def _check_and_parser_gff(self, args_op):
        self._check_gff(args_op.tsss, "tss")
        self._check_gff(args_op.gffs, "gff")
        self._check_gff(args_op.trans, "tran")
        self._check_gff(args_op.utr5s, "utr")
        self._check_gff(args_op.utr3s, "utr")
        self.multiparser.parser_gff(args_op.gffs, None)
        self.multiparser.parser_gff(args_op.tsss, "TSS")
        self.multiparser.combine_gff(args_op.gffs, self.tss_path, None, "TSS")
        self.multiparser.parser_gff(args_op.trans, "transcript")
        self.multiparser.combine_gff(args_op.gffs, self.tran_path,
                                     None, "transcript")
        self.multiparser.parser_gff(args_op.utr5s, "5UTR")
        self.multiparser.combine_gff(args_op.gffs, self.utr5_path,
                                     None, "5UTR")
        self.multiparser.parser_gff(args_op.utr3s, "3UTR")
        self.multiparser.combine_gff(args_op.gffs, self.utr3_path,
                                     None, "3UTR")
        if args_op.terms is not None:
            self._check_gff(args_op.terms, "term")
            self.multiparser.parser_gff(args_op.terms, "term")
            self.multiparser.combine_gff(args_op.gffs, self.term_path,
                                         None, "term")

    def _stat(self, table_path, stat_folder):
        for table in os.listdir(table_path):
            if table.endswith("_operon.csv"):
                filename = "_".join(["stat", table])
                out_stat = os.path.join(stat_folder, filename)
                stat(os.path.join(table_path, table), out_stat)

    def _combine_gff(self, prefixs, args_op):
        '''combine all gff files of features which are associated with operon'''
        for prefix in prefixs:
            out_file = os.path.join(args_op.output_folder, "gffs",
                                    "_".join([prefix, "operon.gff"]))
            print("Generating the gff file of {0}".format(prefix))
            tss = self.helper.get_correct_file(
                    self.tss_path, "_TSS.gff", prefix, None, None)
            tran = self.helper.get_correct_file(
                    self.tran_path, "_transcript.gff", prefix, None, None)
            gff = self.helper.get_correct_file(
                    args_op.gffs, ".gff", prefix, None, None)
            utr5 = self.helper.get_correct_file(
                    self.utr5_path, "_5UTR.gff", prefix, None, None)
            utr3 = self.helper.get_correct_file(
                    self.utr3_path, "_3UTR.gff", prefix, None, None)
            if self.term_path is None:
                term = None
            else:
                term = self.helper.get_correct_file(
                        self.term_path, "_term.gff", prefix, None, None)
            combine_gff(gff, tran, tss, utr5, utr3, term,
                        args_op.tss_fuzzy, args_op.term_fuzzy, out_file)

    def run_operon(self, args_op):
        self._check_and_parser_gff(args_op)
        prefixs = []
        for gff in os.listdir(args_op.gffs):
            if gff.endswith(".gff"):
                prefixs.append(gff.replace(".gff", ""))
        self._detect_operon(prefixs, args_op)
        self._stat(self.table_path, args_op.stat_folder)
        self._combine_gff(prefixs, args_op)
        self.helper.remove_tmp_dir(args_op.gffs)
        self.helper.remove_tmp_dir(args_op.utr3s)
        self.helper.remove_tmp_dir(args_op.utr5s)
        self.helper.remove_tmp_dir(args_op.tsss)
        self.helper.remove_tmp_dir(args_op.trans)
        if args_op.terms is not None:
            self.helper.remove_tmp_dir(args_op.terms)



import os
import sys
from glob import glob
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.parser_wig import WigParser


def load_wigs(out, lib_t, lib_n, lib_f):
    if lib_t and lib_n:
        for index in range(len(lib_t)):
            out.write("load {0}\n".format(
                      os.path.join(os.getcwd(), lib_t[index])))
            out.write("load {0}\n".format(
                      os.path.join(os.getcwd(), lib_n[index])))
    elif lib_t:
        for lib in lib_t:
            out.write("load {0}\n".format(os.path.join(os.getcwd(), lib)))
    elif lib_n:
        for lib in lib_n:
            out.write("load {0}\n".format(os.path.join(os.getcwd(), lib)))
    if lib_f:
        for lib in lib_f:
            out.write("load {0}\n".format(os.path.join(os.getcwd(), lib)))


def set_data_range(out, gff, wigs, strand):
    '''set and print the DataRange'''
    max_range = 0
    for strains in wigs.values():
        for strain, wig_datas in strains.items():
            if strain == gff.seq_id:
                for wig in wig_datas[(gff.start - 1): gff.end]:
                    if max_range < wig.coverage:
                        max_range = wig.coverage
    max_range = int(max_range / 1) + 10
    if strand == "+":
        out.write("setDataRange 0,{0}\n".format(max_range))
    else:
        max_range = max_range * -1
        out.write("setDataRange 0,{0}\n".format(max_range))


def print_batch(args_sc, out, strand, lib_t, lib_n, lib_f, strain):
    '''print the batch file'''
    out.write("new\n")
    out.write("genome {0}\n".format(os.path.join(os.getcwd(), args_sc.fasta)))
    out.write("load {0}\n".format(os.path.join(os.getcwd(), args_sc.main_gff)))
    gff = args_sc.main_gff.split("/")
    out.write("{0} {1}\n".format(args_sc.present, gff[-1]))
    for files in args_sc.side_gffs:
        for filename in glob(files):
            out.write("load {0}\n".format(os.path.join(os.getcwd(), filename)))
            gff = filename.split("/")
            out.write("{0} {1}\n".format(args_sc.present, gff[-1]))
    load_wigs(out, lib_t, lib_n, lib_f)
    out.write("maxPanelHeight {0}\n".format(args_sc.height))
    if strand == "+":
        out.write("snapshotDirectory {0}\n".format(
                      os.path.join(os.getcwd(), args_sc.output_folder,
                                   strain, "forward")))
    else:
        out.write("snapshotDirectory {0}\n".format(
                      os.path.join(os.getcwd(), args_sc.output_folder,
                                   strain, "reverse")))


def import_wig(lib, wigs, strand):
    wig_parser = WigParser()
    for wig in lib:
        wigs[wig] = {}
        strain = ""
        wig_fh = open(wig)
        for entry in wig_parser.parser(wig_fh, strand):
            if strain != entry.strain:
                wigs[wig][entry.strain] = []
                strain = entry.strain
            wigs[wig][strain].append(entry)
        wig_fh.close()


def gen_batch(lib_t, lib_n, lib_f, strand, gffs, out, seq):
    '''generate the batch file'''
    wigs = {}
    if lib_t and lib_n:
        import_wig(lib_t, wigs, strand)
        import_wig(lib_n, wigs, strand)
    elif lib_t:
        import_wig(lib_t, wigs, strand)
    elif lib_n:
        import_wig(lib_n, wigs, strand)
    if lib_f:
        import_wig(lib_f, wigs, strand)
    if strand == "+":
        print("Printing the forward batch files...")
    else:
        print("Printing the reverse batch files...")
    for gff in gffs:
        if gff.seq_id not in seq.keys():
            print("Error: The strain names in fasta file "
                  "and gff file are different!!")
            sys.exit()
        if (gff.start - 200) <= 0:
            start = 1
        else:
            start = gff.start - 200
        if (gff.end + 200) >= len(seq[gff.seq_id]):
            end = len(seq[gff.seq_id])
        else:
            end = gff.end + 200
        out.write("goto {0}:{1}-{2}\n".format(
                  gff.seq_id, start, end))
        set_data_range(out, gff, wigs, strand)
        out.write("snapshot {0}:{1}-{2}.png\n".format(
                  gff.seq_id, gff.start, gff.end))


def get_length(fasta_file):
    '''get sequence information and we can know the length of seq'''
    seq = {}
    with open(fasta_file) as fh:
        for line in fh:
            line = line.strip()
            if line.startswith(">"):
                strain = line[1:]
                seq[strain] = ""
            else:
                seq[strain] = seq[strain] + line
    return seq


def gen_screenshot(args_sc, libs, forward_file, reverse_file, strain):
    '''Generation of screenshot of IGV for reveiwing of user'''
    gffs_f = []
    gffs_r = []
    fh = open(args_sc.main_gff)
    for entry in Gff3Parser().entries(fh):
        if entry.strand == "+":
            gffs_f.append(entry)
        else:
            gffs_r.append(entry)
    gffs_f = sorted(gffs_f, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    gffs_r = sorted(gffs_r, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    out_f = open(forward_file, "w")
    print_batch(args_sc, out_f, "+", libs["ft"],
                libs["fn"], libs["ff"], strain)
    out_r = open(reverse_file, "w")
    print_batch(args_sc, out_r, "-", libs["rt"],
                libs["rn"], libs["rf"], strain)
    seq = get_length(args_sc.fasta)
    gen_batch(libs["ft"], libs["fn"], libs["ff"], "+", gffs_f, out_f, seq)
    gen_batch(libs["rt"], libs["rn"], libs["rf"], "-", gffs_r, out_r, seq)
    fh.close()
    out_f.close()
    out_r.close()



import csv


def import_data(row):
    return{"strain": row[1], "strand": row[2],
           "associate": row[3], "start_seq": int(row[4]),
           "end_seq": int(row[5]), "rfam": row[6], "e": row[7],
           "start_align": int(row[8]), "end_align": int(row[9]),
           "info": "|".join(row[0:6]), "ID": row[0]}


def read_file(ribo_table, rfam_table):
    ribos = []
    rfams = []
    f_h = open(ribo_table, "r")
    for row in csv.reader(f_h, delimiter="\t"):
        if not row[0].startswith("#"):
            ribos.append(import_data(row))
    r_h = open(rfam_table, "r")
    for row in csv.reader(r_h, delimiter="\t"):
        rfams.append({"ID": row[0].strip(), "class": row[1].strip()})
    ribos = sorted(ribos, key=lambda x: (x["strain"], x["start_seq"]))
    f_h.close()
    r_h.close()
    return ribos, rfams


def get_overlap(pre_ribo, ribo, overlap, overlaps):
    if (pre_ribo["strain"] == ribo["strain"]) and \
       (pre_ribo["strand"] == ribo["strand"]) and \
       (pre_ribo["ID"] == ribo["ID"]):
        overlap = True
    if overlap:
        detect = False
        for over in overlaps[ribo["strain"]]:
            if pre_ribo["info"] in over:
                over = over + ";" + ribo["info"]
                detect = True
        if not detect:
            overlaps[ribo["strain"]].append(
                     pre_ribo["info"] + ";" + ribo["info"])


def print_gff(num, ribo, out, stats, strain, feature):
    attribute = ";".join(["=".join(items) for items in [
                          ("ID", "_".join([strain, feature.lower() + str(num)])),
                          ("Name", ribo["rfam_name"]),
                          ("rfam_id", ribo["rfam"]),
                          ("e_value", ribo["e"]),
                          ("method", "infernal_to_Rfam")]])
    out.write("\t".join([str(field) for field in [
              ribo["strain"], "ANNOgesic", feature,
              str(ribo["start_seq"]), str(ribo["end_seq"]),
              ".", ribo["strand"], ".", attribute]]) + "\n")
    stats["total"]["total"] += 1
    stats[strain]["total"] += 1


def import_stat(rfams, ribo, stats, strain):
    for rfam in rfams:
        if ribo["rfam"] == rfam["ID"]:
            ribo["rfam_name"] = rfam["class"]
            if rfam["class"] not in stats["total"].keys():
                stats["total"][rfam["class"]] = 1
            else:
                stats["total"][rfam["class"]] += 1
            if rfam["class"] not in stats[strain].keys():
                stats[strain][rfam["class"]] = 1
            else:
                stats[strain][rfam["class"]] += 1


def print_number(stats, repeat, out, strain, feature):
    out.write("Total number of potential {0} are {1}\n".format(
               feature.replace("_", " "), stats[strain]["total"]))
    out.write("The number of potential {0} which "
              "have overlap region with others are {1}\n".format(
               feature.replace("_", " "), repeat,))
    out.write(feature + "_name\tnumbers\n")
    for type_, num in stats[strain].items():
        if type_ != "total":
            out.write("{0}\t{1}\n".format(type_, num))


def print_stat(stats, out_stat, overlaps, feature):
    out = open(out_stat, "w")
    print_file = False
    repeat = 0
    if len(stats) > 2:
        out.write("All strains:\n")
        print_file = True
        for strain, overs in overlaps.items():
            for over in overs:
                datas = over.split(";")
                repeat = repeat + len(datas)
        print_number(stats, repeat, out, "total", feature)
    for strain, datas in stats.items():
        repeat = 0
        if strain != "total":
            print_file = True
            out.write("{0}:\n".format(strain))
            for over in overlaps[strain]:
                datas = over.split(";")
                repeat = repeat + len(datas)
            print_number(stats, repeat, out, strain, feature)
            print_strain = strain
    if print_file:
        count = 1
        if len(stats) > 2:
            for strain, overs in overlaps.items():
                for over in overs:
                    datas = over.split(";")
                    out.write("\noverlap candidates set {0}:\n".format(count))
                    count += 1
                    for data in datas:
                        out.write("\t{0}\n".format(data))
        else:
            for over in overlaps[print_strain]:
                datas = over.split(";")
                out.write("\noverlap candidates set {0}:\n".format(count))
                count += 1
                for data in datas:
                    out.write("\t{0}\n".format(data))
    out.close()


def stat_and_covert2gff(ribo_table, rfam_table, gff_file, fuzzy, out_stat,
                        feature):
    '''do statistics and print gff file of riboswitch'''
    stats = {}
    overlaps = {}
    pre_strain = ""
    stats["total"] = {"total": 0}
    num = 0
    ribos, rfams = read_file(ribo_table, rfam_table)
    out = open(gff_file, "w")
    out.write("##gff-version 3\n")
    pre_gff = None
    for ribo in ribos:
        overlap = False
        if ribo["strain"] != pre_strain:
            overlaps[ribo["strain"]] = []
            first = True
            strain = ribo["strain"]
            pre_strain = ribo["strain"]
            stats[strain] = {"total": 0}
        if first:
            first = False
            pre_ribo = ribo
        else:
            get_overlap(pre_ribo, ribo, overlap, overlaps)
            pre_ribo = ribo
        if ribo["start_align"] > fuzzy:
            ribo["start_seq"] = ribo["start_seq"] + ribo["start_align"] - fuzzy
        if (ribo["end_seq"] - (ribo["start_seq"] + ribo["end_align"])) > fuzzy:
            ribo["end_seq"] = ribo["start_seq"] + ribo["end_align"] + fuzzy
        import_stat(rfams, ribo, stats, strain)
        if pre_gff is not None:
            if (pre_gff["strain"] == ribo["strain"]) and (
                    pre_gff["strand"] == ribo["strand"]) and (
                    pre_gff["start_seq"] == ribo["start_seq"]) and (
                    pre_gff["end_seq"] == ribo["end_seq"]):
                pre_gff["rfam_name"] = "/".join(
                        [pre_gff["rfam_name"], ribo["rfam_name"]])
                pre_gff["rfam"] = ",".join([pre_gff["rfam"], ribo["rfam"]])
                pre_gff["e"] = ",".join([pre_gff["e"], ribo["e"]])
            else:
                print_gff(num, pre_gff, out, stats, strain, feature)
                num += 1
                pre_gff = ribo
        else:
            pre_gff = ribo
    print_gff(num, pre_gff, out, stats, strain, feature)
    print_stat(stats, out_stat, overlaps, feature)
    out.close()



from annogesiclib.gff3 import Gff3Parser


def read_file(filename):
    datas = []
    f_h = open(filename, "r")
    for entry in Gff3Parser().entries(f_h):
        datas.append(entry)
    datas = sorted(datas, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    f_h.close()
    return datas


def del_attributes(entry):
    '''delete the feature (besides genome annotation) 
    which already has Parent'''
    if (entry.feature == "CDS") or (
            entry.feature == "tRNA") or (
            entry.feature == "rRNA") or (
            entry.feature == "exon"):
        pass
    else:
        if "Parent" in entry.attributes.keys():
            del entry.attributes["Parent"]


def print_file(entry, tran, out):
    if "Parent" in entry.attributes.keys():
        entry.attributes["Parent"] = ",".join([entry.attributes["Parent"],
                                               str(tran.attributes["ID"])])
    else:
        entry.attributes["Parent"] = str(tran.attributes["ID"])
    attributes = {}
    for key, value in entry.attributes.items():
        if (key != "print") and (key != "parent_tran"):
            attributes[key] = value
    attribute_string = ";".join(
        ["=".join(items) for items in attributes.items()])
    out.write("".join([entry.info_without_attributes, "\t",
                       attribute_string, "\n"]))
    entry.attributes["print"] = True


def compare_tran(datas, tran, out):
    '''comare transcript and 5UTR, 3UTR, gene/CDS for merging'''
    for data in datas:
        del_attributes(data)
        if (data.seq_id == tran.seq_id) and (
                data.strand == tran.strand):
            if (data.start >= tran.start) and (
                    data.end <= tran.end):
                print_file(data, tran, out)


def print_rest(datas, out):
    '''print the rest data which is not related to operon'''
    for data in datas:
        if "print" not in data.attributes.keys():
            out.write(data.info + "\n")


def compare_tran_term(term, tran, out, fuzzy_term):
    '''compare transcript and terminator for merging'''
    if (term.seq_id == tran.seq_id) and (
            term.strand == tran.strand):
        if (term.start >= tran.start) and (
                term.end <= tran.end):
            print_file(term, tran, out)
        else:
            if term.strand == "+":
                if ((term.start - fuzzy_term) <= tran.end) and (
                        term.end + fuzzy_term >= tran.end):
                    print_file(term, tran, out)
                elif (term.start <= tran.end) and (
                        term.end >= tran.end):
                    print_file(term, tran, out)
            else:
                if (term.end + fuzzy_term >= tran.start) and (
                        term.start - fuzzy_term <= tran.start):
                    print_file(term, tran, out)
                elif (term.start <= tran.start) and (
                        term.end >= tran.start):
                    print_file(term, tran, out)


def combine_gff(gff_file, ta_file, tss_file, utr5_file, utr3_file,
                term_file, fuzzy_tss, fuzzy_term, out_file):
    '''combine the features which related to operon to 
    form a operon gff file'''
    gffs = read_file(gff_file)
    trans = read_file(ta_file)
    tsss = read_file(tss_file)
    utr5s = read_file(utr5_file)
    utr3s = read_file(utr3_file)
    out = open(out_file, "w")
    out.write("##gff-version 3\n")
    if term_file is not None:
        terms = read_file(term_file)
    for tran in trans:
        out.write(tran.info + "\n")
        for tss in tsss:
            del_attributes(tss)
            if (tss.seq_id == tran.seq_id) and (tss.strand == tran.strand):
                if tss.strand == "+":
                    if ((tss.start + fuzzy_tss) >= tran.start) and (
                            tss.start <= tran.end):
                        print_file(tss, tran, out)
                else:
                    if (tss.start >= tran.start) and (
                            tss.end - fuzzy_tss <= tran.end):
                        print_file(tss, tran, out)
        compare_tran(utr5s, tran, out)
        compare_tran(gffs, tran, out)
        compare_tran(utr3s, tran, out)
        if term_file is not None:
            for term in terms:
                del_attributes(term)
                compare_tran_term(term, tran, out, fuzzy_term)
    print_rest(tsss, out)
    print_rest(utr5s, out)
    print_rest(gffs, out)
    print_rest(utr3s, out)
    if term_file is not None:
        print_rest(terms, out)
    out.close()



import os
import csv
import shutil
from annogesiclib.gff3 import Gff3Parser


def read_gff(gff_file):
    cdss = []
    g_h = open(gff_file)
    for entry in Gff3Parser().entries(g_h):
        if (entry.feature == "CDS") or (
                entry.feature == "tRNA") or (
                entry.feature == "rRNA"):
            cdss.append(entry)
    cdss = sorted(cdss, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    g_h.close()
    return cdss


def check_repeat(tab, strain, strand, start, end):
    if (tab["strain"] == strain) and (
            tab["strand"] == strand):
        if ((tab["start"] <= start) and (
                 tab["end"] >= end)) or (
                (tab["start"] >= start) and (
                 tab["end"] <= end)) or (
                (tab["start"] <= start) and (
                 tab["end"] <= end) and (
                 tab["end"] >= start)) or (
                (tab["start"] >= start) and (
                 tab["start"] <= end) and (
                 tab["end"] >= end)):
            return True


def rbs_overlap(table_file, gff_file):
    tmp_tab = table_file + "_tmp"
    cdss = read_gff(gff_file)
    out = open(tmp_tab, "w")
    fh = open(table_file, "r")
    tables = []
    for row in csv.reader(fh, delimiter='\t'):
        if not row[0].startswith("#"):
            tables.append({"strain": row[1], "strand": row[2],
                           "start": int(row[4]), "end": int(row[5]),
                           "info": "\t".join(row)})
    fh.close()
    for tab in tables:
        overlap = False
        for cds in cdss:
            overlap = check_repeat(tab, cds.seq_id, cds.strand,
                                   cds.start, cds.end)
        for com in tables:
            if tab != com:
                repeat = check_repeat(tab, com["strain"], com["strand"],
                                      com["start"], com["end"])
                if (not overlap):
                    if ((repeat) and (
                            "print" not in tab.keys()) and (
                            "print" not in com.keys())) or (
                            not repeat):
                        overlap = False
                    else:
                        overlap = True
        if not overlap:
            tab["print"] = True
            out.write(tab["info"] + "\n")
    out.close()
    os.remove(table_file)
    shutil.move(tmp_tab, table_file)



import math
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.lib_reader import read_libs, read_wig


def read_gff(input_file):
    datas = []
    gff_parser = Gff3Parser()
    f_h = open(input_file, "r")
    for entry in gff_parser.entries(f_h):
        entry.attributes["print"] = False
        datas.append(entry)
    datas = sorted(datas, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    return datas


def get_coverage(tar, wigs):
    '''get coverage'''
    coverage = 0
    for strain, conds in wigs.items():
        if tar.seq_id == strain:
            for tracks in conds.values():
                for wigs in tracks.values():
                    if coverage < wigs[tar.start - 1]["coverage"]:
                        coverage = wigs[tar.start - 1]["coverage"]
    return coverage


def compare_wig(tars, wig_fs, wig_rs):
    '''get the coverage of TSS for comparison'''
    for tar in tars:
        if tar.strand == "+":
            tar.attributes["coverage"] = get_coverage(tar, wig_fs)
        elif tar.strand == "-":
            tar.attributes["coverage"] = get_coverage(tar, wig_rs)


def stat(tars, refs, cutoff, gene_length, cluster):
    '''do statistics and print it out'''
    stats = {"tp": 0, "fp": 0, "miss": 0, "fp_rate": 0,
             "tp_rate": 0, "miss_rate": 0}
    num_ref = 0
    for ref in refs:
        num_ref += 1
        detect = False
        for tar in tars:
            if (ref.seq_id == tar.seq_id) and (
                    ref.strand == tar.strand) and (
                    float(tar.attributes["coverage"]) >= cutoff) and (
                    tar.start <= int(gene_length)):
                if math.fabs(ref.start - tar.start) <= cluster:
                    stats["tp"] += 1
                    tar.attributes["print"] = True
                    detect = True
        if not detect:
            stats["miss"] += 1
    for tar in tars:
        if (not tar.attributes["print"]) and (
                float(tar.attributes["coverage"]) >= cutoff) and (
                tar.start <= int(gene_length)):
            stats["fp"] += 1
    stats["fp_rate"] = float(stats["fp"]) / float(int(gene_length) - num_ref)
    stats["tp_rate"] = float(stats["tp"]) / float(num_ref)
    stats["miss_rate"] = float(stats["miss"]) / float(num_ref)
    return stats, num_ref


def print_file(tars, cutoff, out_file):
    out = open(out_file, "w")
    for tar in tars:
        if tar.attributes["coverage"] >= cutoff:
            out.write(tar.info + "\n")


def change_best(num_ref, best, stat_value):
    '''scoring function for evaluate the change of TSS candidates'''
    change = False
    if num_ref > 100:
        if best["tp_rate"] - stat_value["tp_rate"] >= 0.1:
            change = False
        else:
            if (best["tp_rate"] <= stat_value["tp_rate"]) and (
                    best["fp_rate"] >= stat_value["fp_rate"]):
                best = stat_value.copy()
                change = True
            elif (stat_value["tp_rate"] - best["tp_rate"] >= 0.01) and (
                    stat_value["fp_rate"] - best["fp_rate"] <= 0.00005):
                best = stat_value.copy()
                change = True
            elif (best["tp_rate"] - stat_value["tp_rate"] <= 0.01) and (
                    best["fp_rate"] - stat_value["fp_rate"] >= 0.00005):
                best = stat_value.copy()
                change = True
    else:
        if best["tp"] - stat_value["tp"] >= 5:
            change = False
        else:
            if (best["tp"] <= stat_value["tp"]) and (
                    best["fp"] >= stat_value["fp"]):
                best = stat_value.copy()
                change = True
            tp_diff = float(best["tp"] - stat_value["tp"])
            if tp_diff > 0:
                if float(best["fp"] - stat_value["fp"]) >= 5 * tp_diff:
                    best = stat_value.copy()
                    change = True
            elif tp_diff < 0:
                tp_diff = tp_diff * -1
                if float(stat_value["fp"] - best["fp"]) <= 5 * tp_diff:
                    best = stat_value.copy()
                    change = True
    return best, change


def filter_low_expression(gff_file, args_tss, wig_f_file,
                          wig_r_file, out_file):
    '''filter the low expressed TSS'''
    tars = read_gff(gff_file)
    refs = read_gff(args_tss.manual_file)
    libs, texs = read_libs(args_tss.input_lib, args_tss.wig_folder)
    wig_fs = read_wig(wig_f_file, "+", args_tss.libs)
    wig_rs = read_wig(wig_r_file, "-", args_tss.libs)
    compare_wig(tars, wig_fs, wig_rs)
    cutoff = 1
    first = True
    while True:
        stat_value, num_ref = stat(tars, refs, cutoff,
                                   args_tss.gene_length, args_tss.cluster)
        if first:
            first = False
            best = stat_value.copy()
            continue
        else:
            best, change = change_best(num_ref, best, stat_value)
            if not change:
                break
        cutoff = cutoff + 0.1
    print_file(tars, cutoff, out_file)
    return cutoff



from annogesiclib.gff3 import Gff3Parser


class FormatFixer(object):
    '''Fix the format which is not fit with ANNOgesic'''

    def _read_gff(self, gff_file, genes, datas, strain):
        gene_num = 0
        fh = open(gff_file, "r")
        for entry in Gff3Parser().entries(fh):
            entry.seq_id = strain
            entry.info_without_attributes = "\t".join([
                str(field) for field in [
                        entry.seq_id, entry.source, entry.feature, entry.start,
                        entry.end, entry.score, entry.strand, entry.phase]])
            datas.append(entry)
            if entry.feature == "gene":
                if "locus_tag" in entry.attributes.keys():
                    name = entry.attributes["locus_tag"]
                if "gene" in entry.attributes.keys():
                    name = entry.attributes["gene"]
                entry.attribute_string = ";".join(["ID=gene" + str(gene_num),
                                                   "Name=" + name,
                                                   entry.attribute_string])
                gene_id = "gene" + str(gene_num)
                entry.attributes["ID"] = gene_id
                genes.append(entry)
                gene_num += 1
        fh.close()

    def fix_ratt(self, gff_file, strain, out_file):
        out = open(out_file, "w")
        out.write("##gff-version 3\n")
        nums = {"cds": 0, "rna": 0, "gene": 0}
        genes = []
        datas = []
        check_parent = False
        self._read_gff(gff_file, genes, datas, strain)
        check_parent = False
        for data in datas:
            if data.feature == "gene":
                data = genes[nums["gene"]]
                nums["gene"] += 1
            elif (data.feature == "rRNA") or \
                 (data.feature == "tRNA"):
                name = data.attributes["locus_tag"]
                data.attribute_string = ";".join([
                    "ID=rna" + str(nums["rna"]),
                    "Name=" + name, data.attribute_string])
                nums["rna"] += 1
            elif data.feature == "CDS":
                if "protein_id" in data.attributes.keys():
                    name = data.attributes["protein_id"]
                for gene in genes:
                    if ((gene.start <= data.start) and (
                            gene.end >= data.end)) or (
                            gene.attributes["locus_tag"] ==
                            data.attributes["locus_tag"]):
                        data.attribute_string = ";".join([
                            "ID=cds" + str(nums["cds"]), "Name=" + name,
                            "Parent=" + gene.attributes["ID"],
                            data.attribute_string])
                        check_parent = True
                        break
                if check_parent:
                    check_parent = False
                    pass
                else:
                    data.attribute_string = ";".join([
                        "ID=cds" + str(nums["cds"]),
                        "Name=" + name, data.attribute_string])
                nums["cds"] += 1
            if "group" in data.attributes.keys():
                ref_f = open(gff_file, "r")
                for ref in Gff3Parser().entries(ref_f):
                    if "group" in ref.attributes.keys():
                        if (data.attributes["group"] ==
                                ref.attributes["group"]):
                            if (data.strand != ref.strand):
                                data.strand = ref.strand
                            break
                ref_f.close()
            out.write("\t".join([data.info_without_attributes,
                                 data.attribute_string]) + "\n")
        out.close()

    def fix_rnaplex(self, rnaplex_file, out_file):
        out = open(out_file, "w")
        with open(rnaplex_file, "r") as f_h:
            for line in f_h:
                line = line.strip()
                if line != ("Error during initialization of "
                            "the duplex in duplexfold_XS"):
                    out.write(line + "\n")
        out.close()

    def fix_emboss(self, input_file, out_file):
        out = open(out_file, "w")
        with open(input_file, "r") as f_h:
            for line in f_h:
                line = line.strip()
                if line.startswith(">"):
                    out.write(line[:-2] + "\n")
                else:
                    out.write(line + "\n")
        out.close()



import os
from subprocess import call
from annogesiclib.gen_svg import gen_svg
from annogesiclib.helper import Helper


class ColorPNG(object):

    def _convert_svg(self, imagemagick_path, out_path, screenshot, svg_file):
        call([imagemagick_path,
              os.path.join(out_path, screenshot),
              os.path.join(out_path, svg_file)])

    def _convert_png(self, imagemagick_path, out_path, screenshot, png_file):
        call([imagemagick_path, "-background", "none",
              os.path.join(out_path, screenshot),
              os.path.join(out_path, png_file)])

    def generate_color_png(self, track_num, out_folder, imagemagick_path):
        '''generation of color png based on tracks'''
        out_folder = os.path.join(out_folder, "screenshots")
        for strain in os.listdir(out_folder):
            if os.path.isdir(os.path.join(out_folder, strain)):
                for strand in ["forward", "reverse"]:
                    print("Running for {0}_{1}".format(strain, strand))
                    out_path = os.path.join(out_folder, strain, strand)
                    # convert original png to svg and give color on it.
                    for screenshot in os.listdir(out_path):
                        if screenshot.endswith(".png"):
                            print("Converting {0} to svg and "
                                  "Painting tracks now".format(
                                      screenshot))
                            svg_file = screenshot.replace(".png", ".svg")
                            self._convert_svg(imagemagick_path, out_path,
                                              screenshot, svg_file)
                            with open(os.path.join(
                                      out_path, svg_file), "r") as f_h:
                                for line in f_h:
                                    line = line.strip()
                                    if line.startswith("<svg"):
                                        line = line.split(" ")
                                        height = line[-1].split("=")[-1][1:-2]
                                        width = line[1].split("=")[-1][1:-1]
                                        break
                            gen_svg(os.path.join(out_path, screenshot),
                                    track_num, height, width)
                    # convert to png file again
                    for screenshot in os.listdir(out_path):
                        if screenshot.endswith(".svg"):
                            print("Converting {0} to png now...".format(
                                  screenshot))
                            png_file = screenshot.replace(".svg", ".png")
                            print(os.path.join(out_path, png_file))
                            self._convert_png(imagemagick_path, out_path,
                                              screenshot, png_file)
                    Helper().remove_all_content(out_path, ".svg", "file")



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='ANNOgesic',
    version='0.5.10',
    packages=['annogesiclib'],
    author='Sung-Huan Yu',
    author_email='sung-huan.yu@uni-wuerzburg.de',
    description='A bacterial transcriptome annotation pipeline based on RNA-Seq',
    url='',
    install_requires=[
        "biopython >= 1.65",
        "matplotlib >= 1.5.0",
        "numpy >= 1.9.2",
        "networkx >= 1.9.1"
    ],
    scripts=['bin/annogesic'],
    license='ISC License (ISCL)',
    classifiers=[
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ]
)



import os
import math
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.parser_wig import WigParser
from annogesiclib.helper import Helper


def get_primary_locus_tag(tss):
    tsss = []
    tss_types = tss.attributes["type"].split(",")
    tss_locus_tags = tss.attributes["associated_gene"].split(",")
    tss_utr_lengths = tss.attributes["utr_length"].split(",")
    index = 0
    for tss_type in tss_types:
        if "Primary" in tss_type:
            tsss.append({"locus": tss_locus_tags[index],
                         "utr": int(tss_utr_lengths[index].split("_")[1]),
                         "type": tss_type})
        index += 1
    return tsss


def detect_coverage(wigs, tss, ref):
    for strain, tracks in wigs.items():
        if strain == tss.seq_id:
            tss_cover = 0
            ref_cover = 0
            for wig in tracks.values():
                if ((tss.start + 1) <= len(wig)) and (
                        (ref.start + 1) <= len(wig)):
                    if tss.strand == "+":
                        diff_t = (wig[tss.start - 1]["coverage"] -
                                  wig[tss.start - 2]["coverage"])
                        diff_r = (wig[ref.start - 1]["coverage"] -
                                  wig[ref.start - 2]["coverage"])
                    else:
                        diff_t = (wig[tss.start - 1]["coverage"] -
                                  wig[tss.start]["coverage"])
                        diff_r = (wig[ref.start - 1]["coverage"] -
                                  wig[ref.start]["coverage"])
                    tss_cover = tss_cover + diff_t
                    ref_cover = ref_cover + diff_r
    return tss_cover, ref_cover


def fix_attributes(tss, tss_entry):
    '''change the primary TSS to secondary TSS'''
    index = 0
    genes = tss.attributes["associated_gene"].split(",")
    utrs = tss.attributes["utr_length"].split(",")
    types = tss.attributes["type"].split(",")
    for gene in genes:
        if gene == tss_entry["locus"]:
            utrs[index] = utrs[index].replace("Primary", "Secondary")
            types[index] = types[index].replace("Primary", "Secondary")
        index += 1
    tss.attributes["utr_length"] = ",".join(utrs)
    tss.attributes["type"] = ",".join(types)


def del_repeat(tsss):
    '''delete the repeat TSS'''
    for tss in tsss:
        types = tss.attributes["type"].split(",")
        utrs = tss.attributes["utr_length"].split(",")
        genes = tss.attributes["associated_gene"].split(",")
        detect = {"pri": False, "sec": False}
        index = 0
        finals = {"types": [], "utrs": [], "genes": []}
        for type_ in types:
            if (type_ == "Primary") and (detect["pri"] is False):
                detect["pri"] = True
                pri_utr = int(utrs[index].split("_")[1])
                real_index = index
            elif (type_ == "Primary") and (detect["pri"] is True):
                compare_utr = int(utrs[index].split("_")[1])
                if compare_utr < pri_utr:
                    pri_utr = compare_utr
                    real_index = index
            elif (type_ == "Secondary") and (detect["sec"] is False):
                detect["sec"] = True
                sec_utr = int(utrs[index].split("_")[1])
                real_index2 = index
            elif (type_ == "Secondary") and (detect["sec"] is True):
                compare_utr = int(utrs[index].split("_")[1])
                if compare_utr < sec_utr:
                    sec_utr = compare_utr
                    real_index2 = index
            elif (type_ == "Antisense") or \
                 (type_ == "Internal") or \
                 (type_ == "Orphan"):
                finals["types"].append(types[index])
                finals["utrs"].append(utrs[index])
                finals["genes"].append(genes[index])
            index += 1
        if detect["pri"] is True:
            finals["types"].append(types[real_index])
            finals["utrs"].append(utrs[real_index])
            finals["genes"].append(genes[real_index])
        else:
            if detect["sec"] is True:
                finals["types"].append(types[real_index2])
                finals["utrs"].append(utrs[real_index2])
                finals["genes"].append(genes[real_index2])
        tss.attributes["type"] = ",".join(finals["types"])
        tss.attributes["utr_length"] = ",".join(finals["utrs"])
        tss.attributes["associated_gene"] = ",".join(finals["genes"])


def fix_primary_type(tsss, wigs_f, wigs_r):
    '''if one gene is associated with multiple TSSs, it will 
    check the coverage and assign the low expressed TSS to be 
    secondary TSS'''
    for tss in tsss:
        if ("Primary" in tss.attributes["type"]):
            tss_entrys = get_primary_locus_tag(tss)
            for ref in tsss:
                if (ref.seq_id == tss.seq_id) and \
                   (ref.strand == tss.strand) and \
                   (ref.start == tss.start):
                    pass
                else:
                    if ("Primary" in ref.attributes["type"]):
                        ref_entrys = get_primary_locus_tag(ref)
                        for tss_entry in tss_entrys:
                            for ref_entry in ref_entrys:
                                if (tss_entry["locus"] ==
                                    ref_entry["locus"]) and (
                                        tss_entry["type"] == "Primary") and (
                                        ref_entry["type"] == "Primary") and (
                                        tss.seq_id == ref.seq_id):
                                    if tss.strand == "+":
                                        tss_cover, ref_cover = detect_coverage(
                                            wigs_f, tss, ref)
                                    else:
                                        tss_cover, ref_cover = detect_coverage(
                                            wigs_r, tss, ref)
                                    if tss_cover < ref_cover:
                                        fix_attributes(tss, tss_entry)
                                    elif tss_cover > ref_cover:
                                        fix_attributes(ref, ref_entry)
                                    elif tss_cover == ref_cover:
                                        if (tss_entry["utr"] <
                                                ref_entry["utr"]):
                                            fix_attributes(ref, ref_entry)
                                        elif (tss_entry["utr"] >
                                                ref_entry["utr"]):
                                            fix_attributes(tss, tss_entry)
    del_repeat(tsss)
    return tsss


def define_attributes(tss):
    string = []
    for key, value in tss.attributes.items():
        if key != "print":
            if key != "ID":
                string.append("=".join([key, value]))
            elif key == "Name":
                string.append("=".join([key, str(tss.start) + tss.strand]))
    return ";".join(string)


def remove_primary(tss, tss_entry):
    final_types = []
    final_utrs = []
    final_genes = []
    tss_dict = tss_entry[1]
    types = tss_dict["type"].split(",")
    utrs = tss_dict["utr_length"].split(",")
    genes = tss_dict["associated_gene"].split(",")
    index = 0
    for type_ in types:
        if type_ != "Primary":
            final_types.append(type_)
            final_utrs.append(utrs[index])
            final_genes.append(genes[index])
        index += 1
    tss_dict = {"Name": "TSS_" + str(tss.start) + tss.strand,
                "type": ",".join(final_types),
                "utr_length": ",".join(final_utrs),
                "associated_gene": ",".join(final_genes)}
    tss_string = ";".join(["=".join(["utr_length", tss_dict["utr_length"]]),
                           "=".join(["associated_gene",
                                     tss_dict["associated_gene"]]),
                           "=".join(["type", tss_dict["type"]]),
                           "=".join(["Name", tss_dict["Name"]])])
    return [tss_string, tss_dict]


def import_to_tss(tss_type, cds_pos, tss, locus_tag, tss_entry):
    if cds_pos == "NA":
        utr = "_".join([tss_type, "NA"])
    else:
        utr = "_".join([tss_type, str(int(math.fabs(cds_pos - tss.start)))])
    if len(tss_entry) != 0:
        tss_dict = tss_entry[1]
        tss_dict_types = tss_dict["type"].split(",")
        tss_dict_utrs = tss_dict["utr_length"].split(",")
        tss_dict_tags = tss_dict["associated_gene"].split(",")
        if tss_type == "Primary" and ("Primary" in tss_dict["type"]):
            index = 0
            for tss_dict_type in tss_dict_types:
                if "Primary" in tss_dict_type:
                    utr_length = tss_dict_utrs[index].split("_")
                    if math.fabs(cds_pos - tss.start) < int(utr_length[1]):
                        tss_dict_utrs[index] = utr
                        tss_dict_tags[index] = locus_tag
                index += 1
        else:
            tss_dict_types.append(tss_type)
            tss_dict_utrs.append(utr)
            tss_dict_tags.append(locus_tag)
        tss_dict = {"Name": "TSS_" + str(tss.start) + tss.strand,
                    "type": ",".join(tss_dict_types),
                    "utr_length": ",".join(tss_dict_utrs),
                    "associated_gene": ",".join(tss_dict_tags)}
    else:
        tss_dict = {"Name": "TSS_" + str(tss.start) + tss.strand,
                    "type": tss_type,
                    "utr_length": utr,
                    "associated_gene": locus_tag}
    tss_string = ";".join(["=".join(["utr_length", tss_dict["utr_length"]]),
                           "=".join(["associated_gene",
                                     tss_dict["associated_gene"]]),
                           "=".join(["type", tss_dict["type"]]),
                           "=".join(["Name", tss_dict["Name"]])])
    return (tss_string, tss_dict)


def same_strand_tss_gene(gene, tss, anti_ends, gene_ends, checks, tss_entry):
    '''check the TSS and gene which are at the same strand'''
    if is_primary(gene.start, gene.end, tss.start, tss.strand):
        locus_tag = gene.attributes["locus_tag"]
        if tss.strand == "+":
            if ((anti_ends["reverse"] != -1) and (
                    anti_ends["reverse"] - gene.start) > 0) or (
                    anti_ends["reverse"] == -1):
                tss_entry = import_to_tss("Primary", gene.start, tss,
                                          locus_tag, tss_entry)
                checks["orphan"] = False
                gene_ends["forward"] = gene.start
            elif (anti_ends["reverse"] != -1) and (
                    (anti_ends["reverse"] - gene.start) < 0):
                if (checks["int_anti"] is True) or (
                        (tss.start - anti_ends["reverse"]) > 0):
                    tss_entry = import_to_tss("Primary", gene.start, tss,
                                              locus_tag, tss_entry)
                    checks["orphan"] = False
                    gene_ends["forward"] = gene.start
        else:
            if ((anti_ends["forward"] != -1) and (
                    gene.end - anti_ends["forward"]) > 0) or (
                    anti_ends["forward"] == -1):
                tss_entry = import_to_tss("Primary", gene.end, tss,
                                          locus_tag, tss_entry)
                checks["orphan"] = False
                gene_ends["reverse"] = gene.end
    if is_internal(gene.start, gene.end, tss.start, tss.strand):
        locus_tag = gene.attributes["locus_tag"]
        tss_entry = import_to_tss("Internal", "NA", tss, locus_tag, tss_entry)
        checks["orphan"] = False
    return tss_entry


def diff_strand_tss_gene(gene, tss, anti_ends, gene_ends, checks, tss_entry):
    '''check the TSS and gene which are at the same strand'''
    if is_antisense(gene.start, gene.end, tss.start, tss.strand):
        checks["int_anti"] = False
        if tss.strand == "-":
            anti_ends["forward"] = gene.start
            if (gene_ends["reverse"] != -1) and (
                    (gene.start - gene_ends["reverse"]) > 0):
                if is_internal(gene.start, gene.end, tss.start, tss.strand):
                    pass
                else:
                    if (tss.start - gene.end) > 0:
                        tss_entry = remove_primary(tss, tss_entry)
        else:
            anti_ends["reverse"] = gene.end
            if is_internal(gene.start, gene.end, tss.start, tss.strand):
                checks["int_anti"] = True
            if (gene_ends["forward"] != -1) and (
                    (gene.start - gene_ends["forward"]) > 0):
                if (gene.start - tss.start) > 0:
                    tss_entry = remove_primary(tss, tss_entry)
        locus_tag = gene.attributes["locus_tag"]
        tss_entry = import_to_tss("Antisense", "NA", tss, locus_tag, tss_entry)
        checks["orphan"] = False
    return tss_entry


def compare_tss_gene(tss, genes):
    '''compare TSS and gene to find the relation'''
    tss_entry = []
    gene_ends = {"forward": -1, "reverse": -1}
    anti_ends = {"forward": -1, "reverse": -1}
    checks = {"orphan": True, "int_anti": None}
    for gene in genes:
        if gene.strand == tss.strand:
            tss_entry = same_strand_tss_gene(gene, tss, anti_ends,
                                             gene_ends, checks, tss_entry)
        else:
            tss_entry = diff_strand_tss_gene(gene, tss, anti_ends,
                                             gene_ends, checks, tss_entry)
    if checks["orphan"]:
        tss_entry = import_to_tss("Orphan", "NA", tss, "NA", tss_entry)
    return tss_entry


def is_primary(cds_start, cds_end, tss_pos, strand):
    if strand == "+":
        if (is_utr(cds_start, tss_pos, 300) and (cds_start >= tss_pos)):
            return True
    else:
        if (is_utr(tss_pos, cds_end, 300) and (cds_end <= tss_pos)):
            return True


def is_internal(cds_start, cds_end, tss_pos, strand):
    if ((cds_start < tss_pos) and (cds_end > tss_pos)) or (
            (strand == "+") and (tss_pos == cds_end)) or (
            (strand == "-") and (tss_pos == cds_start)):
        return True


def is_antisense(cds_start, cds_end, tss_pos, strand):
    if ((is_utr(cds_start, tss_pos, 100)) and (cds_start >= tss_pos)) or (
            (is_utr(tss_pos, cds_end, 100)) and (cds_end <= tss_pos)) or (
             is_internal(cds_start, cds_end, tss_pos, strand)):
        return True


def is_utr(pos1, pos2, length):
    if (pos1 - pos2 <= length):
        return True


def print_all_unique(out, overlap_num, nums):
    if ((nums["tss_p"] != 0) or (overlap_num != 0)) and (
            (nums["tss_m"] != 0) or (overlap_num != 0)):
        out.write("the number of overlap between "
                  "TSSpredator and manual = {0} ".format(
                      overlap_num))
        out.write("(overlap of all TSSpredator = {0}, ".format(
                  float(overlap_num) /
                  (float(nums["tss_p"]) + float(overlap_num))))
        out.write("overlap of all manual = {0})\n".format(
                  float(overlap_num) /
                  (float(nums["tss_m"]) + float(overlap_num))))
        out.write("the number of unique in TSSpredator = {0} ({1})\n".format(
                  nums["tss_p"], float(nums["tss_p"]) /
                  (float(nums["tss_p"]) + float(overlap_num))))
        out.write("the number of unique in manual = {0} ({1})\n".format(
                  nums["tss_m"], float(nums["tss_m"]) /
                  (float(nums["tss_m"]) + float(overlap_num))))
    else:
        out.write("No TSS candidates which be predicted by TSSpredator.")


def print_stat(num_strain, stat_file, overlap_num, nums):
    if len(num_strain) != 0:
        out = open(stat_file, "w")
        if len(num_strain.keys()) == 1:
            print_all_unique(out, overlap_num, nums)
        else:
            out.write("All strains: \n")
            print_all_unique(out, overlap_num, nums)
            for strain in num_strain.keys():
                if (num_strain[strain]["tsspredator"] == 0) and \
                   (num_strain[strain]["overlap"] == 0):
                    perc_tsspredator = "NA"
                    perc_tsspredator_uni = "NA"
                else:
                    perc_tsspredator = str(
                        float(num_strain[strain]["overlap"]) / (
                            float(num_strain[strain]["tsspredator"]) +
                            float(num_strain[strain]["overlap"])))
                    perc_tsspredator_uni = str(
                        float(num_strain[strain]["tsspredator"]) / (
                            float(num_strain[strain]["tsspredator"]) +
                            float(num_strain[strain]["overlap"])))
                if (num_strain[strain]["manual"] == 0) and (
                        num_strain[strain]["overlap"] == 0):
                    perc_manual = "NA"
                    perc_manual_uni = "NA"
                else:
                    perc_manual = str(
                        float(num_strain[strain]["overlap"]) / (
                            float(num_strain[strain]["manual"]) +
                            float(num_strain[strain]["overlap"])))
                    perc_manual_uni = str(
                        float(num_strain[strain]["manual"]) / (
                            float(num_strain[strain]["manual"]) +
                            float(num_strain[strain]["overlap"])))
                out.write(strain + ": \n")
                out.write("the number of overlap between "
                          "TSSpredator and manual = {0} ".format(
                              num_strain[strain]["overlap"]))
                out.write("(overlap of all TSSpredator = {0}, ".format(
                          perc_tsspredator))
                out.write("overlap of all manual = {0})\n".format(perc_manual))
                out.write("the number of unique in "
                          "TSSpredator = {0} ({1})\n".format(
                              num_strain[strain]["tsspredator"],
                              perc_tsspredator_uni))
                out.write("the number of unique in manual = {0} ({1})\n".format(
                          num_strain[strain]["manual"], perc_manual_uni))


def read_wig(filename, strand):
    wigs = {}
    wig_parser = WigParser()
    if filename:
        wig_fh = open(filename)
        for entry in wig_parser.parser(wig_fh, strand):
            if entry.strain not in wigs.keys():
                strain = entry.strain
                wigs[strain] = {}
            if entry.track not in wigs[strain].keys():
                wigs[strain][entry.track] = []
            wigs[strain][entry.track].append({
                 "pos": entry.pos, "coverage": entry.coverage,
                 "strand": entry.strand})
        wig_fh.close()
    return wigs


def read_gff(tss_predict_file, tss_manual_file, gff_file, lengths):
    tsss = {"tsss_p": [], "tsss_m": [], "merge": []}
    cdss = []
    genes = []
    gff_parser = Gff3Parser()
    tssp_fh = open(tss_predict_file, "r")
    tssm_fh = open(tss_manual_file, "r")
    g_f = open(gff_file, "r")
    for entry in gff_parser.entries(tssp_fh):
        entry.attributes["print"] = False
        tsss["tsss_p"].append(entry)
    tssp_fh.close()
    tsss["tsss_p"] = sorted(tsss["tsss_p"], key=lambda k: (k.seq_id, k.start,
                                                           k.end, k.strand))
    for entry in gff_parser.entries(tssm_fh):
        if (entry.seq_id in lengths.keys()) or ("all" in lengths.keys()):
            entry.attributes["print"] = False
            entry.attributes["libs"] = "manual"
            entry.attributes["method"] = "manual"
            tsss["tsss_m"].append(entry)
    tssm_fh.close()
    tsss["tsss_m"] = sorted(tsss["tsss_m"], key=lambda k: (k.seq_id, k.start,
                                                           k.end, k.strand))
    for entry in gff_parser.entries(g_f):
        if (entry.feature == "CDS") or (
                entry.feature == "rRNA") or (
                entry.feature == "tRNA"):
            cdss.append(entry)
        if entry.feature == "gene":
            genes.append(entry)
    g_f.close()
    cdss = sorted(cdss, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    genes = sorted(genes, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    return tsss, cdss, genes


def merge_libs(input_libs, wig_folder, program):
    if "merge_forward.wig" in os.listdir(os.getcwd()):
        os.remove("merge_forward.wig")
    if "merge_reverse.wig" in os.listdir(os.getcwd()):
        os.remove("merge_reverse.wig")
    if program == "TSS":
        type_ = "tex"
    elif program == "processing":
        type_ = "notex"
    for lib in input_libs:
        datas = lib.split(":")
        if (datas[1] == type_) and (datas[4] == "+"):
            Helper().merge_file(os.path.join(wig_folder, datas[0]),
                                os.path.join(os.getcwd(), "merge_forward.wig"))
        elif (datas[1] == type_) and (datas[4] == "-"):
            Helper().merge_file(os.path.join(wig_folder, datas[0]),
                                os.path.join(os.getcwd(), "merge_reverse.wig"))


def check_overlap(overlap, pre_tss, nums, length, num_strain, overlap_num,
                  tss_m, tss_p, tsss, pre_pos, cdss, genes):
    '''find the TSS which be detected in manual detection and TSSpredator'''
    if overlap:
        if pre_tss:
            pre_tss.attributes["print"] = True
            tss = pre_tss
        else:
            tss = tss_p
        tss.attribute_string = define_attributes(tss)
        tss.attributes["method"] = "TSSpredator,manual"
        if (not length) or \
           (tss.start <= int(length)):
            num_strain[tss.seq_id]["overlap"] += 1
            if (pre_pos != -1):
                if (tss.start - pre_pos != 0):
                    tsss["merge"].append(tss)
                    nums["tss"] += 1
                    overlap_num += 1
                else:
                    overlap_num += 1
            else:
                tsss["merge"].append(tss)
                nums["tss"] += 1
                overlap_num += 1
        overlap = False
        pre_pos = tss.start
    else:
        if tss_m.seq_id == tss_p.seq_id:
            tss_entry = compare_tss_gene(tss_m, genes)
            tss_m.attributes = tss_entry[1]
            tss_m.attribute_string = tss_entry[0]
            tss_m.attributes["method"] = "manual"
            tsss["merge"].append(tss_m)
            if (not length) or \
               (tss_m.start <= int(length)):
                num_strain[tss_m.seq_id]["manual"] += 1
                nums["tss_m"] += 1
                nums["tss"] += 1
    return (overlap, pre_pos, overlap_num)


def intersection(tsss, cluster, nums, lengths, cdss, genes, seqs):
    '''compare the predicted TSS and manual TSS'''
    num_strain = {}
    overlap = False
    overlap_num = 0
    pre_pos = -1
    start = False
    length = None
    for tss_m in tsss["tsss_m"]:
        pre_tss = None
        start = False
        if "all" in lengths.keys():
            length = seqs[tss_m.seq_id]
        else:
            if lengths[tss_m.seq_id] == "all":
                length = seqs[tss_m.seq_id]
            else:
                length = lengths[tss_m.seq_id]
        for tss_p in tsss["tsss_p"]:
            start = True
            if (tss_p.strand == tss_m.strand) and \
               (tss_p.seq_id == tss_m.seq_id):
                if (tss_p.start == tss_m.start):
                    tss_p.attributes["print"] = True
                    overlap = True
                    pre_tss = None
                    break
                elif (math.fabs(tss_p.start - tss_m.start) <= cluster):
                    overlap = True
                    pre_tss = tss_p
        if start:
            if tss_p.seq_id not in num_strain.keys():
                num_strain[tss_p.seq_id] = {"overlap": 0, "tsspredator": 0,
                                            "manual": 0}
            datas = check_overlap(overlap, pre_tss, nums, length, num_strain,
                                  overlap_num, tss_m, tss_p, tsss, pre_pos,
                                  cdss, genes)
            overlap = datas[0]
            pre_pos = datas[1]
            overlap_num = datas[2]
    if (start) or (len(tsss["tsss_m"]) == 0):
        for tss_p in tsss["tsss_p"]:
            run = False
            if not tss_p.attributes["print"]:
                tss_p.attribute_string = define_attributes(tss_p)
                tsss["merge"].append(tss_p)
                if (length is None):
                    run = True
                else:
                   if (tss_p.start <= int(length)):
                       run = True
                if run and (tss_p.seq_id in num_strain):
                    num_strain[tss_p.seq_id]["tsspredator"] += 1
                    nums["tss"] += 1
                    nums["tss_p"] += 1
    return overlap_num, num_strain


def print_file(final_tsss, program, out_gff):
    num_final = 0
    out = open(out_gff, "w")
    out.write("##gff-version 3\n")
    for tss in final_tsss:
        if "print" in tss.attributes.keys():
            del tss.attributes["print"]
        tss.attributes["ID"] = "_".join([
            tss.seq_id, program.lower() + str(num_final)])
        num_final += 1
        if program == "TSS":
            strand = Helper().get_strand_name(tss.strand)
            tss.attributes["Name"] = "TSS:" + "_".join(
                                              [str(tss.start), strand])
        else:
            strand = Helper().get_strand_name(tss.strand)
            tss.attributes["Name"] = "processing:" + "_".join(
                                              [str(tss.start), strand])
        tss.attribute_string = ";".join(
            ["=".join(items) for items in tss.attributes.items()])
        out.write("\t".join([str(field) for field in [
                             tss.seq_id, "ANNOgesic", tss.feature, tss.start,
                             tss.end, tss.score, tss.strand, tss.phase,
                             tss.attribute_string]]) + "\n")


def read_seq(seq_file):
    seqs = {}
    with open(seq_file) as fh:
        for line in fh:
            line = line.strip()
            if line.startswith(">"):
                strain = line[1:]
                seqs[strain] = 0
            else:
                seqs[strain] = seqs[strain] + len(line)
    return seqs

def merge_manual_predict_tss(tss_predict_file, stat_file, out_gff,
                             gff_file, args_tss, manual, seq_file):
    '''merge the manual detected TSS and TSSpredator predicted TSS'''
    nums = {"tss_p": 0, "tss_m": 0, "tss": 0}
    merge_libs(args_tss.libs, args_tss.wig_folder, args_tss.program)
    wigs_f = read_wig("merge_forward.wig", "+")
    wigs_r = read_wig("merge_reverse.wig", "-")
    seqs = read_seq(seq_file)
    tsss, cdss, genes, = read_gff(tss_predict_file, manual,
                                  gff_file, args_tss.strain_lengths)
    overlap_num, num_strain = intersection(tsss, args_tss.cluster, nums,
                                           args_tss.strain_lengths,
                                           cdss, genes, seqs)
    sort_tsss = sorted(tsss["merge"], key=lambda k: (k.seq_id, k.start,
                                                     k.end, k.strand))
    final_tsss = fix_primary_type(sort_tsss, wigs_f, wigs_r)
    print_file(final_tsss, args_tss.program, out_gff)
    print_stat(num_strain, stat_file, overlap_num, nums)
